package com.maroontress.gcovparser;

import com.maroontress.gcovparser.gcda.DataRecord;
import com.maroontress.gcovparser.gcda.FunctionDataRecord;
import com.maroontress.gcovparser.gcno.FunctionGraphRecord;
import com.maroontress.gcovparser.gcno.NoteRecord;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Collection;
import java.util.TreeMap;

/**
   gcnoեѡ̤ݻޤ

   @param <T> ؿ
*/
public abstract class AbstractNote<T extends AbstractFunctionGraph> {

    /** gcnoեΥꥸǤ */
    private Origin origin;

    /** gcnoեΥСֹǤ */
    private int version;

    /**
       gcnoեΥॹפǤgcdaեƱȤƤ
       뤳Ȥǧ뤿˻Ѥޤ
    */
    private int stamp;

    /** ؿդȤμ̻ҤΥޥåפǤ */
    private TreeMap<Integer, T> map;

    /** ץμ¹ԲǤ */
    private int runs;

    /** ץοǤ */
    private int programs;

    /**
       󥹥󥹤ޤ

       @param path gcnoեΥѥ
    */
    public AbstractNote(final String path) {
	origin = new Origin(path);
    }

    /**
       ؿե쥳ɤؿդޤ

       @param e ؿե쥳
       @return ؿ
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    protected abstract T createFunctionGraph(
	FunctionGraphRecord e) throws CorruptedFileException;

    /**
       ꥸޤ

       @return ꥸ
    */
    public final Origin getOrigin() {
	return origin;
    }

    /**
       gcnoեΥСޤ

       @return gcnoեΥС
    */
    protected final int getVersion() {
	return version;
    }

    /**
       gcnoեΥॹפޤ

       @return gcnoեΥॹ
    */
    protected final int getStamp() {
	return stamp;
    }

    /**
       ץμ¹Բޤ

       @return ץμ¹Բ
    */
    protected final int getRuns() {
	return runs;
    }

    /**
       ץοޤ

       @return ץο
    */
    protected final int getPrograms() {
	return programs;
    }

    /**
       ؿդΥ쥯ޤ

       @return ؿդΥ쥯
    */
    public final Collection<T> getFunctionGraphCollection() {
	return map.values();
    }

    /**
       Ρȥ쥳ɤꤷޤ

       @param rec Ρȥ쥳
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private void setNoteRecord(final NoteRecord rec)
	throws CorruptedFileException {
	version = rec.getVersion();
	stamp = rec.getStamp();
	map = new TreeMap<Integer, T>();

	FunctionGraphRecord[] list = rec.getList();
	for (FunctionGraphRecord e : list) {
	    T fg = createFunctionGraph(e);
	    map.put(fg.getId(), fg);
	}
    }

    /**
       gcnoեѡơΡȤޤͥޥå
       Τǡ2GХȤĶեϰޤ

       @throws IOException ϥ顼
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    protected final void parseNote() throws IOException,
	CorruptedFileException {
	File noteFile = origin.getNoteFile();
	RandomAccessFile file = new RandomAccessFile(noteFile, "r");
	FileChannel ch = file.getChannel();
	try {
	    ByteBuffer bb = ch.map(FileChannel.MapMode.READ_ONLY,
				   0, ch.size());
	    NoteRecord noteRecord = new NoteRecord(bb);
	    setNoteRecord(noteRecord);
	} finally {
	    file.close();
	}
    }

    /**
       ǡ쥳ɤꤷޤ

       @param rec ǡ쥳
       @param file gcdaե
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private void setDataRecord(final DataRecord rec, final File file)
	throws CorruptedFileException {
	String path = origin.getNoteFile().getPath();
	if (version != rec.getVersion()) {
	    throw new CorruptedFileException(path + ": version mismatch.");
	}
	if (stamp != rec.getStamp()) {
	    throw new CorruptedFileException(path + ": timestamp mismatch.");
	}
	if (origin.getNoteFile().lastModified() > file.lastModified()) {
	    System.err.printf(
		"%s: warning: gcno file is newer than gcda file.%n", path);
	}
	FunctionDataRecord[] list = rec.getList();
	for (FunctionDataRecord e : list) {
	    int id = e.getId();
	    T g = map.get(id);
	    if (g == null) {
		System.err.printf(
		    "%s: warning: unknown function id '%d'.%n", path, id);
		continue;
	    }
	    g.setFunctionDataRecord(e);
	}
	runs = rec.getObjectSummary().getRuns();
	programs = rec.getProgramSummaries().length;
    }

    /**
       gcdaեѡơΡȤ˥󥿤ɲäޤ
       ͥޥåפΤǡ2GХȤĶեϰޤ

       @throws IOException ϥ顼
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    protected final void parseData() throws IOException,
	CorruptedFileException {
	File dataFile = origin.getDataFile();
	RandomAccessFile file = new RandomAccessFile(dataFile, "r");
	FileChannel ch = file.getChannel();
	try {
	    ByteBuffer bb = ch.map(FileChannel.MapMode.READ_ONLY,
				   0, ch.size());
	    DataRecord dataRecord = new DataRecord(bb);
	    setDataRecord(dataRecord, dataFile);
	} finally {
	    file.close();
	}
    }
}
